from time import sleep

import troi
from random import randint, shuffle

import requests

from troi import Recording
from troi.plist import plist
from troi import TARGET_NUMBER_OF_RECORDINGS


class LBRadioTagRecordingElement(troi.Element):

    NUM_RECORDINGS_TO_COLLECT = TARGET_NUMBER_OF_RECORDINGS * 4

    def __init__(self,
                 tags,
                 operator="and",
                 mode="easy",
                 include_similar_tags=True):
        troi.Element.__init__(self)
        self.tags = tags
        self.operator = operator
        self.mode = mode
        self.include_similar_tags = include_similar_tags

    def inputs(self):
        return []

    def outputs(self):
        return [Recording]

    def fetch_similar_tags(self, tag):
        """
            Fetch similar tags from LB
        """

        while True:
            r = requests.post("https://labs.api.listenbrainz.org/tag-similarity/json", json=[{"tag": tag}])
            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code == 404:
                return plist()

            if r.status_code != 200:
                raise RuntimeError(f"Cannot fetch similar tags. {r.text}")

            break

        return plist(r.json())

    def select_recordings(self):

        msgs = []
        start, stop = { "easy": (66, 95), "medium": (33, 66), "hard": (1, 33) }[self.mode]
        sim_start, sim_stop = { "easy": (0, 0), "medium": (50, 100), "hard": (10, 50) }[self.mode]
        num_similar_tags_to_include = { "easy": 0, "medium": 1, "hard": 2 }[self.mode] 

        recordings = self.recording_search_by_tag.search(self.tags, self.operator, start, stop,
                                                         self.NUM_RECORDINGS_TO_COLLECT)
        if not recordings:
            return [], ["Could not find any recordings for tag search '%s', ignoring." % (",".join(self.tags)) ] 

        if len(self.tags) == 1 and self.include_similar_tags:
            similar_tags = self.fetch_similar_tags(self.tags[0])

            if len(similar_tags[sim_start:sim_stop]) > 2:
                while True:
                    selected_tags = similar_tags.random_item(count=2)
                    if selected_tags[0] == selected_tags[1]:
                        continue

                    break
                similar_tags = selected_tags
            else:
                similar_tags = similar_tags[sim_start:sim_stop]

            similar_tags = [tag["similar_tag"] for tag in similar_tags]
            if len(similar_tags) > 0:
                for i in range(num_similar_tags_to_include):
                    sim_tag_data = self.recording_search_by_tag.search(
                        (self.tags[0], similar_tags[i]), "AND", start, stop,
                        self.NUM_RECORDINGS_TO_COLLECT)
                    if len(sim_tag_data) > self.NUM_RECORDINGS_TO_COLLECT:
                        sim_tag_data = sim_tag_data.random_item( start, stop, self.NUM_RECORDINGS_TO_COLLECT)

                    recordings.extend(sim_tag_data)

                if num_similar_tags_to_include > 1:
                    msgs = [
                        f"""tag: using seed tag '{self.tags[0]}' and similar tags '{"', '".join(similar_tags)}'."""
                    ]
                else:
                    msgs = [
                        f"""tag: using seed tag '{self.tags[0]}' and similar tag '{similar_tags[0]}'."""
                    ]
            else:
                msgs = [f"""tag: using only seed tag '{self.tags[0]}'."""]


        recordings = list(recordings)
        shuffle(recordings)
        return recordings, msgs


    def read(self, entities):

        self.recording_search_by_tag = self.patch.get_service(
            "recording-search-by-tag")

        self.local_storage["data_cache"]["element-descriptions"].append(
            f'tag{"" if len(self.tags) == 1 else "s"} {", ".join(self.tags)}')

        recordings, feedback = self.select_recordings()
        for msg in feedback:
            self.local_storage["user_feedback"].append(msg)

        return recordings



from time import sleep
from random import randint, shuffle
from uuid import UUID

import requests
from urllib.parse import quote

import troi.patch
import troi.filters
import troi.listenbrainz.feedback
import troi.listenbrainz.listens
import troi.listenbrainz.recs
import troi.musicbrainz.recording_lookup
from troi.playlist import PlaylistMakerElement
from troi.parse_prompt import PromptParser, ParseError, TIME_RANGES
from troi.patches.lb_radio_classes.artist import LBRadioArtistRecordingElement
from troi.patches.lb_radio_classes.blend import InterleaveRecordingsElement, WeighAndBlendRecordingsElement
from troi.patches.lb_radio_classes.collection import LBRadioCollectionRecordingElement
from troi.patches.lb_radio_classes.playlist import LBRadioPlaylistRecordingElement
from troi.patches.lb_radio_classes.tag import LBRadioTagRecordingElement
from troi.patches.lb_radio_classes.stats import LBRadioStatsRecordingElement
from troi.patches.lb_radio_classes.recs import LBRadioRecommendationRecordingElement
from troi.patches.lb_radio_classes.country import LBRadioCountryRecordingElement
from troi import TARGET_NUMBER_OF_RECORDINGS, Playlist
from troi.utils import interleave


class LBRadioPatch(troi.patch.Patch):
    """
       ListenBrainz radio experimentation.
    """

    # If the user specifies no time_range, default to this one
    DEFAULT_TIME_RANGE = "month"

    def __init__(self, args):
        self.artist_mbids = []
        self.mode = None

        # Remember, the create function for this class will be called in the super() init.
        super().__init__(args)

    @staticmethod
    def inputs():
        """
        Generate a playlist from one or more Artist MBIDs

        \b
        MODE which mode to generate playlists in. must be one of easy, mediumedium, hard
        PROMPT is the LB radio prompt. See troi/parse_prompt.py for details.
        """
        return [{
            "type": "argument",
            "args": ["mode"],
            "kwargs": {
                "required": True,
                "nargs": 1
            }
        }, {
            "type": "argument",
            "args": ["prompt"],
            "kwargs": {
                "required": True,
                "nargs": 1
            }
        }]

    @staticmethod
    def outputs():
        return [Playlist]

    @staticmethod
    def slug():
        return "lb-radio"

    @staticmethod
    def description():
        return "Given an LB radio prompt, generate a playlist for that prompt."

    def lookup_artist(self, artist_name):
        """ Fetch artist names for validation purposes """

        if isinstance(artist_name, UUID):
            return self.lookup_artist_from_mbid(artist_name)

        err_msg = f"Artist {artist_name} could not be looked up. Please use exact spelling."

        while True:
            r = requests.get( f"https://musicbrainz.org/ws/2/artist?query={quote(artist_name)}&fmt=json")
            if r.status_code == 404:
                raise RuntimeError(err_msg)

            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code != 200:
                raise RuntimeError( f"Could not resolve artist name {artist_name}. Error {r.status_code} {r.text}")

            break

        data = r.json()
        try:
            fetched_name = data["artists"][0]["name"]
            mbid = data["artists"][0]["id"]
        except (IndexError, KeyError):
            raise RuntimeError(err_msg)

        if fetched_name.lower() == artist_name.lower():
            return fetched_name, mbid

        raise RuntimeError(err_msg)

    def lookup_artist_from_mbid(self, artist_mbid):
        """ Fetch artist names for validation purposes """

        while True:
            r = requests.get(f"https://musicbrainz.org/ws/2/artist/%s?fmt=json" % str(artist_mbid))
            if r.status_code == 404:
                raise RuntimeError(f"Could not resolve artist mbid {artist_mbid}. Error {r.status_code} {r.text}")

            if r.status_code in (429, 503):
                sleep(2)
                continue

            if r.status_code != 200:
                raise RuntimeError(f"Could not resolve artist name {artist_mbid}. Error {r.status_code} {r.text}")

            break

        return r.json()["name"], artist_mbid

    def create(self, inputs):
        self.prompt = inputs["prompt"]
        self.mode = inputs["mode"]

        # First parse the prompt
        pp = PromptParser()
        try:
            prompt_elements = pp.parse(self.prompt)
        except ParseError as err:
            raise RuntimeError(f"cannot parse prompt: '{err}'")

        if self.mode not in ("easy", "medium", "hard"):
            raise RuntimeError(
                "Argument mode must be one one easy, medium or hard.")

        # Lookup artist names embedded in the prompt
        artist_names = {}
        for element in prompt_elements:
            if element["entity"] == "artist":
                name, mbid = self.lookup_artist(element["values"][0])
                element["values"][0] = mbid
                artist_names[mbid] = name

        # Save descriptions to local storage
        self.local_storage["data_cache"] = {
            "element-descriptions": [],
            "prompt": self.prompt
        }
        self.local_storage["user_feedback"] = []

        weights = [e["weight"] for e in prompt_elements]

        # Now create the graph, based on the mode and the entities desired
        elements = []
        for element in prompt_elements:
            mode = self.mode
            if "easy" in element["opts"]:
                mode = "easy"
            elif "medium" in element["opts"]:
                mode = "medium"
            if "hard" in element["opts"]:
                mode = "hard"

            # Determine percent ranges based on mode -- this will likely need further tweaking
            if mode == "easy":
                start, stop = 0, 33
            elif self.mode == "medium":
                start, stop = 33, 66
            else:
                start, stop = 66, 100
            self.local_storage["modes"] = {
                "easy": (0, 33),
                "medium": (33, 66),
                "hard": (66, 100)
            }

            if element["entity"] == "artist":
                include_sim = False if "nosim" in element["opts"] else True
                source = LBRadioArtistRecordingElement(
                    element["values"][0],
                    artist_name=artist_names[element["values"][0]],
                    mode=mode,
                    include_similar_artists=include_sim)

            if element["entity"] == "tag":
                include_sim = False if "nosim" in element["opts"] else True
                operator = "or" if "or" in element["opts"] else "and"
                source = LBRadioTagRecordingElement(
                    [ t.lower() for t in element["values"]],
                    mode=mode,
                    operator=operator,
                    include_similar_tags=include_sim)

            if element["entity"] == "country":
                if isinstance(element["values"][0], UUID):
                    source = LBRadioCountryRecordingElement(
                        mode, 
                        area_mbid=element["values"][0])
                else:
                    source = LBRadioCountryRecordingElement(
                        mode, 
                        area_name=element["values"][0])

            if element["entity"] == "collection":
                source = LBRadioCollectionRecordingElement(
                    element["values"][0], mode=mode)

            if element["entity"] == "playlist":
                source = LBRadioPlaylistRecordingElement(element["values"][0],
                                                         mode=mode)

            if element["entity"] == "stats":
                if len(element["opts"]) == 0:
                    element["opts"].append(self.DEFAULT_TIME_RANGE)
                if len(element["values"]) == 0:
                    raise RuntimeError(
                        "user name cannot be blank for user entity. (at least not yet -- soon it will be)"
                    )
                if len(element["opts"]) != 1:
                    raise RuntimeError(
                        "The user entity needs to define one time range option."
                    )
                source = LBRadioStatsRecordingElement(
                    element["values"][0],
                    mode=mode,
                    time_range=element["opts"][0])

            if element["entity"] == "recs":
                if len(element["values"]) == 0:
                    raise RuntimeError(
                        "user name cannot be blank for user entity. (at least not yet -- soon it will be)"
                    )
                if len(element["opts"]) == 0:
                    listened = "all"
                else:
                    listened = element["opts"][0]
                source = LBRadioRecommendationRecordingElement(
                    element["values"][0], mode=mode, listened=listened)

            recs_lookup = troi.musicbrainz.recording_lookup.RecordingLookupElement(
            )
            recs_lookup.set_sources(source)

            hate_filter = troi.filters.HatedRecordingsFilterElement()
            hate_filter.set_sources(recs_lookup)

            elements.append(hate_filter)

        # Finish the pipeline with the element that blends and weighs the streams
        blend = WeighAndBlendRecordingsElement(weights, max_num_recordings=100, max_artist_occurrence=3)
        blend.set_sources(elements)

        pl_maker = PlaylistMakerElement(
            patch_slug=self.slug(),
            max_num_recordings=TARGET_NUMBER_OF_RECORDINGS)
        pl_maker.set_sources(blend)

        return pl_maker

    def post_process(self):
        """ 
            Take the information saved in local_storage and create proper playlist names and descriptions.
        """

        prompt = self.local_storage["data_cache"]["prompt"]
        names = ", ".join(
            self.local_storage["data_cache"]["element-descriptions"])
        name = f"LB Radio for {names} on {self.mode} mode"
        desc = "Experimental ListenBrainz radio using %s mode, which was generated from this prompt: '%s'" % (
            self.mode, prompt)
        self.local_storage["_playlist_name"] = name
        self.local_storage["_playlist_desc"] = desc



from collections import defaultdict
from datetime import datetime, timedelta
from typing import Optional
from time import sleep

import requests

from troi import Element, Recording


class RecentListensTimestampLookup(Element):
    """
        Element to look up the time when a user last listened given recordings in past X days.
        Note that the element is stateful and caches the recent listens lookup results!

        Timestamps are stored in the listenbrainz dict, with key name "latest_listened_at".

        :param user_name: The ListenBrainz user for whome to fetch recent listen timestamps.
        :param auth_token: a ListenBrainz auth token
        :param days: The number of days to check.
    """

    def __init__(self, user_name, days: int, auth_token=None):
        super().__init__()
        self.user_name = user_name
        self.days = days
        self.auth_token = auth_token
        self.index: Optional[dict[str, int]] = None

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def _fetch_recent_listens_index(self):
        """ Return an index of recording mbids as key and the latest listened_at time of the corresponding
        recording as values.
        """
        index = defaultdict(int)

        min_dt = datetime.now() + timedelta(days=-self.days)
        min_ts = int(min_dt.timestamp())
        while True:
            headers = {"Authorization": f"Token {self.auth_token}"} if self.auth_token else {}
            response = requests.get(
                f"https://api.listenbrainz.org/1/user/{self.user_name}/listens",
                params={"min_ts": min_ts, "count": 100},
                headers=headers
            )
            if response.status_code == 429:
                sleep(2)
                continue

            response.raise_for_status()
            data = response.json()["payload"]
            if len(data["listens"]) == 0:
                break

            for listen in data["listens"]:
                listened_at = listen["listened_at"]

                # add both user submitted mbid and mapped mbid to index

                additional_info = listen["track_metadata"]["additional_info"]
                user_submitted_mbid = additional_info.get("recording_mbid")
                if user_submitted_mbid:
                    index[user_submitted_mbid] = max(index[user_submitted_mbid], listened_at)

                mbid_mapping = listen["track_metadata"].get("mbid_mapping")
                if mbid_mapping:
                    mapped_mbid = mbid_mapping.get("recording_mbid")
                    index[mapped_mbid] = max(index[mapped_mbid], listened_at)

            min_ts = data["listens"][0]["listened_at"]

        return index

    def _get_latest_listened_ts(self, r: Recording):
        # check latest listened timestamp for original mbid
        ts1 = self.index.get(r.mbid)

        # check latest listened timstamp for canonical mbid
        ts2 = None
        if r.listenbrainz is not None and r.listenbrainz.get("canonical_recording_mbid"):
            canonical_mbid = r.listenbrainz.get("canonical_recording_mbid")
            ts2 = self.index.get(canonical_mbid)

        # if we have timestamp for both canonical and normal mbid take the maximum of two else whichever
        # one is available. if none is available continue ahead
        if ts1 and ts2:
            return max(ts1, ts2)
        elif ts1:
            return ts1
        elif ts2:
            return ts2
        else:
            return None

    def read(self, inputs):
        recordings = inputs[0]
        if not recordings:
            return []

        if self.index is None:
            self.index = self._fetch_recent_listens_index()

        for r in recordings:
            ts = self._get_latest_listened_ts(r)
            if ts is None:
                continue

            latest_listened_at = datetime.fromtimestamp(ts).replace(tzinfo=None)
            r.listenbrainz["latest_listened_at"] = latest_listened_at

        return recordings



import json
import unittest
import unittest.mock

import troi
import troi.musicbrainz.mbid_mapping

return_data = [
    {
        "artist_credit_arg": "morcheeba",
        "artist_credit_id": 963,
        "artist_credit_name": "Morcheeba",
        "index": 0,
        "recording_arg": "trigger hippie",
        "recording_mbid": "97e69767-5d34-4c97-b36a-f3b2b1ef9dae",
        "recording_name": "Trigger Hippie",
        "release_mbid": "9db51cd6-38f6-3b42-8ad5-559963d68f35",
        "release_name": "Who Can You Trust?",
        "artist_mbids": ["067102ea-9519-4622-9077-57ca4164cfbb"]
    }
]


def mocked_requests_post(*args, **kwargs):
    class MockResponse:
        def __init__(self, json_data, status_code):
            self.json_data = json_data
            self.status_code = status_code

        def json(self):
            return self.json_data

    e = troi.musicbrainz.mbid_mapping.MBIDMappingLookupElement()
    if args[0] == e.SERVER_URL:
        return MockResponse(return_data, 200)

    return MockResponse(None, 404)


class TestMBIDMapping(unittest.TestCase):

    @unittest.mock.patch('requests.post', side_effect=mocked_requests_post)
    def test_read(self, req):

        e = troi.musicbrainz.mbid_mapping.MBIDMappingLookupElement()

        r = [ troi.Recording("trigger hippie", artist_credit=troi.ArtistCredit(name="morcheeba")) ]
        entities = e.read([r])
        req.assert_called_with(e.SERVER_URL, json=[{'artist_credit_name': 'morcheeba', 'recording_name': 'trigger hippie'}])

        assert len(entities) == 1
        assert entities[0].artist_credit.artist_credit_id == 963
        assert entities[0].artist_credit.name == "Morcheeba"
        assert entities[0].artist_credit.artists[0].mbid == "067102ea-9519-4622-9077-57ca4164cfbb"
        assert entities[0].release.mbid == "9db51cd6-38f6-3b42-8ad5-559963d68f35"
        assert entities[0].release.name == "Who Can You Trust?"
        assert entities[0].mbid == "97e69767-5d34-4c97-b36a-f3b2b1ef9dae"
        assert entities[0].name == "Trigger Hippie"

    @unittest.mock.patch('requests.post')
    def test_read_remove_unmatched(self, req):

        mock = unittest.mock.MagicMock()
        mock.status_code = 200
        mock.text = "{}"
        req.return_value = mock
        e = troi.musicbrainz.mbid_mapping.MBIDMappingLookupElement(True)

        r = [ troi.Recording("track not found", artist_credit=troi.ArtistCredit(name="artist not found")) ]
        entities = e.read([r])
        assert len(entities) == 0



import json

from troi.playlist import _deserialize_from_jspf, PlaylistElement


def read_jspf_playlist(jspf_file):
    """
        Read a JSPF playlist from disk.
    """

    with open(jspf_file, "r") as f:
        jspf = f.read()

    playlist = _deserialize_from_jspf(json.loads(jspf))
    playlist_element = PlaylistElement()
    playlist_element.playlists = [ playlist ]

    return playlist_element


def write_jspf_playlist(jspf_file, jspf):
    """
        Write a JSPF playlist to disk.
    """

    with open(jspf_file, "w") as f:
        f.write(json.dumps(jspf.get_jspf()))


def write_m3u_playlist(file_name, playlist_element):
    """
       Given a Troi playlist, write an m3u playlist to disk.
    """

    playlist = playlist_element.playlists[0]
    with open(file_name, "w") as m3u:
        m3u.write("#EXTM3U\n")
        m3u.write("#EXTENC: UTF-8\n")
        m3u.write("#PLAYLIST %s\n" % playlist.name)
        for rec in playlist.recordings:
            if rec is None:
                continue
            if rec.duration is None:
                duration = 0
            else: 
                duration = rec.duration / 1000
            m3u.write("#EXTINF %d,%s\n" % (duration, rec.name))
            m3u.write(rec.musicbrainz["filename"] + "\n")



import logging
from abc import abstractmethod
from collections import namedtuple
from enum import IntEnum
import os
import datetime
from mutagen import MutagenError
from pathlib import Path
import sys
from types import SimpleNamespace
from uuid import UUID

import peewee
from tqdm import tqdm

from troi.content_resolver.model.database import db, setup_db
from troi.content_resolver.model.recording import Recording, RecordingMetadata, FileIdType
from troi.content_resolver.model.unresolved_recording import UnresolvedRecording
from troi.content_resolver.model.tag import Tag, RecordingTag
from troi.content_resolver.model.directory import Directory
from troi.content_resolver.formats import mp3, m4a, flac, ogg_opus, ogg_vorbis, wma

from troi.content_resolver.utils import existing_dirs

logger = logging.getLogger(__name__)

SUPPORTED_FORMATS = (
    flac,
    m4a,
    mp3,
    ogg_opus,
    ogg_vorbis,
    wma,
)

ALL_EXTENSIONS = set()
EXTENSION_HANDLER = dict()
for fmt in SUPPORTED_FORMATS:
    ALL_EXTENSIONS.update(fmt.EXTENSIONS)
    for ext in fmt.EXTENSIONS:
        EXTENSION_HANDLER[ext] = fmt


class Status(IntEnum):
    NOCHANGE = 0
    ADD = 1
    UPDATE = 2
    ERROR = 255


STATUSMSG = {
    Status.NOCHANGE: '',
    Status.ADD: 'add',
    Status.UPDATE: 'update',
    Status.ERROR: 'error',
}

StatusDetails = namedtuple('StatusDetails', ('recording_name', 'artist_name', 'release_name'))
StatusData = namedtuple('StatusData', ('status', 'filenumber', 'details'))


def match_extensions(filepath, extensions):
    return Path(filepath).suffix.lower() in extensions


class ScanCounters:
    total = 0
    status = {s: 0 for s in Status}
    files = 0
    audio_files = 0
    directories = 0
    updated_directories = 0
    skipped_directories = 0

    def dry_run_stats(self):
        return ("Found {c.audio_files} audio file(s) among {c.files} file(s) in "
                "{c.directories} directorie(s) ({c.skipped_directories} skipped)").format(c=self)

    def _stats(self):
        yield "Checked {count} tracks:".format(count=self.total)
        yield "{count:>8} tracks not changed since last run".format(count=self.status[Status.NOCHANGE])
        yield "{count:>8} tracks added".format(count=self.status[Status.ADD])
        yield "{count:>8} tracks updated".format(count=self.status[Status.UPDATE])
        yield "{count:>8} tracks could not be read".format(count=self.status[Status.ERROR])

        if self.total != sum(self.status.values()):
            yield "And for some reason these numbers don't add up to the total number of tracks. Hmmm."

        if self.updated_directories:
            yield "{count} directory entries updated.".format(count=self.updated_directories)

    def stats(self):
        return "\n".join(self._stats())


class Database:
    '''
    Keep a database with metadata for a collection of local music files.
    '''

    def __init__(self, db_file, quiet):
        self.db_file = db_file
        self.fuzzy_index = None
        self.forced_scan = False
        self.quiet = quiet

    def create(self):
        """
            Create the database. Can be run again to create tables that have been recently added to the code,
            but don't exist in the DB yet.
        """
        try:
            db_dir = os.path.dirname(os.path.realpath(self.db_file))
            os.makedirs(db_dir, exist_ok=True)
            setup_db(self.db_file)
            db.connect()
            db.create_tables((
                Recording,
                RecordingMetadata,
                Tag,
                RecordingTag,
                UnresolvedRecording,
                Directory,
            ))
        except Exception as e:
            logger.error("Failed to create db file %r: %s" % (self.db_file, e))

    def open(self):
        """
            Open the database file and connect to the db.
        """
        try:
            setup_db(self.db_file)
            db.connect()
        except peewee.OperationalError:
            logger.error("Cannot open database index file: '%s'" % self.db_file)
            sys.exit(-1)

    def close(self):
        """ Close the db."""
        db.close()

    def scan(self, music_dirs, chunksize=100, force=False):
        """
            Scan music directories and add tracks to sqlite.
        """
        if not music_dirs:
            logger.error("No directory to scan")
            return

        self.forced_scan = force

        self.music_dirs = tuple(sorted(set(existing_dirs(music_dirs))))
        if not self.music_dirs:
            logger.error("No valid directories to scan")
            return

        self.chunksize = chunksize

        # Keep some stats
        self.counters = ScanCounters()
        self.skip_dirs = set()

        if not self.quiet:
            logger.info("Check collection size...")
            logger.info("Counting candidates in %s ..." % ", ".join(self.music_dirs))
        self.traverse(dry_run=True)
        if not self.quiet:
            logger.info(self.counters.dry_run_stats())

        if not self.quiet:
            with tqdm(total=self.counters.audio_files) as self.progress_bar:
                logger.info("Scanning ...")
                self.traverse()
        else:
            self.traverse()

        self.close()
        if not self.quiet:
            logger.info(self.counters.stats())

    def traverse(self, dry_run=False):
        """
            This function searches for audio files and descends into sub directories
        """
        seen = set()
        changed_dirs = []
        if dry_run:
            self.counters.directories = 0
            self.counters.audio_files = 0
        filenumber = 0
        self.chunk = dict()

        for topdir in self.music_dirs:
            if dry_run:
                self.counters.directories += 1

            for root, dirs, files in os.walk(topdir):
                root = os.path.realpath(root)
                dir_mtime = self.dir_has_changed(root)

                if dry_run:
                    self.counters.directories += len(dirs)
                    if not self.forced_scan and dir_mtime is False:
                        self.skip_dirs.add(root)

                if not self.forced_scan and root in self.skip_dirs:
                    self.counters.skipped_directories += 1
                    continue

                for name in files:
                    file_path = os.path.realpath(os.path.join(root, name))
                    if file_path in seen:
                        continue
                    seen.add(file_path)
                    if os.path.isfile(file_path) and match_extensions(file_path, ALL_EXTENSIONS):
                        filenumber += 1
                        if not dry_run:
                            self.add(file_path, filenumber)
                            if filenumber % self.chunksize == 0:
                                self.process_chunk()

                if not dry_run and dir_mtime is not False:
                    # add changed directory info after it was explored
                    changed_dirs.append({'dir_path': root, 'mtime': dir_mtime})

        if not dry_run:
            self.process_chunk()
            # update directory table after everything was processed.
            # It reduces the risk of issues in case of interruption or crash
            if changed_dirs:
                with db.atomic():
                    Directory.insert_many(changed_dirs).on_conflict_replace().execute()
                    self.counters.updated_directories = len(changed_dirs)
        else:
            self.counters.files = len(seen)
            self.counters.audio_files = filenumber

    def dir_has_changed(self, dir_path):
        """ Returns directory mtime if it changed since last run, or False"""
        try:
            stats = os.stat(dir_path)
            mtime = datetime.datetime.fromtimestamp(stats[8])
            directory = Directory.get_or_none(Directory.dir_path == dir_path)
            if directory is None or directory.mtime != mtime:
                return mtime
        except Exception as e:
            logger.error("Can't stat dir %r: %s" % (dir_path, e))
        return False

    def read_metadata(self, file_path, mtime):
        """
            Read metadata from audio file
            On error, returns None, error msg
            On success, returns metadata dict, StatusDetails
        """
        data = None
        try:
            base, extension = os.path.splitext(file_path)
            handler = EXTENSION_HANDLER[extension]
            tags = handler.READER(file_path)
            mdata = handler.get_metadata(tags)
            if mdata is not None:
                data = {
                    "artist_mbid": self.convert_to_uuid(mdata["artist_mbid"]),
                    "artist_name": mdata["artist_name"],
                    "disc_num": mdata["disc_num"],
                    "file_id": file_path,
                    "file_id_type": FileIdType.FILE_PATH,
                    "mtime": mtime,
                    "recording_mbid": self.convert_to_uuid(mdata["recording_mbid"]),
                    "recording_name": mdata["recording_name"],
                    "release_mbid": self.convert_to_uuid(mdata["release_mbid"]),
                    "release_name": mdata["release_name"],
                    "track_num": mdata["track_num"],
                }
                details = StatusDetails(
                    recording_name=data['recording_name'],
                    artist_name=data['artist_name'],
                    release_name=data['release_name'],
                )
                return data, details
            else:
                return None, "Not enough metadata from file %r" % file_path
        except MutagenError as e:
            return None, "Cannot read metadata from file %r: %s" % (file_path, e)
        except Exception as e:
            return None, "Failed to read audio file %r: %s" % (file_path, e)

    def iterate_chunk(self, chunk):
        """
            For all items in the chunk, read metadata and yield resulting data (or None),
            and matching details (status, filenumber, and details (or error string))
        """
        for file_path, chunkitemdata in chunk.items():
            data, details = self.read_metadata(file_path, chunkitemdata.mtime)
            if data is not None:
                status = Status.UPDATE if chunkitemdata.is_update else Status.ADD
            else:
                status = Status.ERROR

            yield data, StatusData(status, chunkitemdata.filenumber, details)

    def read_metadata_and_add(self, chunk):
        """
            Read the metadata from supported files and then add the
            recording to the DB.
        """
        statuses = list()
        datas = list()

        for data, statusdata in self.iterate_chunk(chunk):
            statuses.append(statusdata)
            if data is not None:
                datas.append(data)

        if datas:
            with db.atomic():
                result = Recording.insert_many(datas).on_conflict_replace().execute()

        return statuses

    def convert_to_uuid(self, value):
        """
            Convert the given string to a UUID or return None if not a valid UUID.
        """

        if value is not None:
            try:
                return UUID(value)
            except ValueError:
                return None
        return None

    def fmtdetails(self, statusdata):
        """
            Format progress message
        """
        s = "%-8s %5.1f%% " % (STATUSMSG[statusdata.status], 100 * statusdata.filenumber / self.counters.audio_files)
        try:
            s += " %-30s %-30s %-30s" % (
                (statusdata.details.recording_name or "")[:29],
                (statusdata.details.artist_name or "")[:29],
                (statusdata.details.release_name or "")[:29],
            )
        except:
            # details can be a string
            s += str(statusdata.details)
        return s

    def update_status(self, statusdata):
        """
            Update status counter and display matching progress
        """
        self.counters.status[statusdata.status] += 1
        if not self.quiet:
            self.progress_bar.write(self.fmtdetails(statusdata))

    def add(self, file_path, audio_file_count):
        """
            Given a file, check to see if we already have it and if we do,
            if it has changed since the last time we read it. If it is new
            or has been changed, update in the DB.
        """

        # update the progress bar
        if not self.quiet:
            self.progress_bar.update(1)

        self.counters.total += 1

        # Check to see if the file in question has changed since the last time
        # we looked at it.
        try:
            stats = os.stat(file_path)
            mtime = datetime.datetime.fromtimestamp(stats[8])
            chunkitemdata = SimpleNamespace(mtime=mtime, filenumber=audio_file_count, is_update=False)
            self.chunk[file_path] = chunkitemdata
        except Exception as e:
            details = "Can't stat file %r: %s" % (file_path, e)
            statusdata = StatusData(Status.ERROR, audio_file_count, details)
            self.update_status(statusdata)

    def process_chunk(self):
        """
            Process current chunk
        """
        statuses = list()

        # find existing recordings and compare modification time
        if not self.forced_scan:
            for recording in Recording.select().where(Recording.file_id.in_(tuple(self.chunk))):
                if recording.mtime == self.chunk[recording.file_id].mtime:
                    # file didn't change since last time, skip it
                    statusdata = StatusData(
                        Status.NOCHANGE, self.chunk[recording.file_id].filenumber,
                        StatusDetails(
                            recording_name=recording.recording_name,
                            artist_name=recording.artist_name,
                            release_name=recording.release_name,
                        ))
                    statuses.append(statusdata)
                    # unchanged files are deleted from chunk
                    del self.chunk[recording.file_id]
                else:
                    #  mark existing data for update
                    self.chunk[recording.file_id].is_update = True

        if self.chunk:
            # add or update metadata for remaining files in the chunk
            statuses += self.read_metadata_and_add(self.chunk)
            # reset chunk
            self.chunk = dict()

        for statusdata in sorted(statuses, key=lambda s: s.filenumber):
            self.update_status(statusdata)

    def database_cleanup(self, dry_run):
        """
        Look for missing files and directory entries and remove them from the DB.
        """
        PathId = namedtuple('PathId', ('path', 'id'))

        recordings = tuple(
            PathId(r.file_id, r.id)
            for r in Recording.select(Recording.file_id, Recording.id).where(Recording.file_id_type == FileIdType.FILE_PATH)
            if not os.path.isfile(r.file_id))
        directories = tuple(
            PathId(d.dir_path, d.id) for d in Directory.select(Directory.dir_path, Directory.id) if not os.path.isdir(d.dir_path))

        if not recordings and not directories:
            logger.error("No cleanup needed.")
            return

        for elem in sorted(recordings + directories):
            logger.info("RM %s" % elem.path)

        logger.info("%d recordings and %d directory entries to remove from database" % (len(recordings), len(directories)))
        if not dry_run:
            with db.atomic():
                ids = tuple(r.id for r in recordings)
                query = Recording.delete().where(Recording.id.in_(ids))
                count = query.execute()
                logger.info("%d recordings removed" % count)
                ids = tuple(d.id for d in directories)
                query = Directory.delete().where(Directory.id.in_(ids))
                count = query.execute()
                logger.info("%d directory entries removed" % count)
            logger.info("Vacuuming database...")
            db.execute_sql('VACUUM')
            logger.info("Done.")
        else:
            logger.info("Use command cleanup --remove to actually remove those.")

    def metadata_sanity_check(self, include_subsonic=False):
        """
        Run a sanity check on the DB to see if data is missing that is required for LB Radio to work.
        """

        num_recordings = db.execute_sql("SELECT COUNT(*) FROM recording").fetchone()[0]
        num_metadata = db.execute_sql("SELECT COUNT(*) FROM recording_metadata").fetchone()[0]
        num_file_path = Recording.select(peewee.fn.Count(
            Recording.file_id_type).alias('count')).where(Recording.file_id_type == FileIdType.FILE_PATH)[0].count
        num_subsonic = Recording.select(peewee.fn.Count(
            Recording.file_id_type).alias('count')).where(Recording.file_id_type == FileIdType.SUBSONIC_ID)[0].count

        if num_metadata == 0:
            logger.info("sanity check: You have not downloaded metadata for your collection. Run the metadata command.")
        elif num_metadata < num_recordings // 2:
            logger.info("sanity check: Only %d of your %d recordings have metadata information available."
                        " Run the metdata command." % (num_metadata, num_recordings))

        if include_subsonic:
            if num_subsonic == 0 and include_subsonic:
                logger.info("sanity check: You have not matched your collection against the collection in subsonic."
                            " Run the subsonic command.")
            elif num_subsonic < num_recordings // 2:
                logger.info("sanity check: Only %d of your %d recordings have subsonic matches."
                            " Run the subsonic command." % (num_subsonic, num_recordings))



import copy
from collections import defaultdict
from time import sleep

import requests

from troi import Element, Recording, PipelineError, DEVELOPMENT_SERVER_URL


class RelatedArtistCreditsElement(Element):
    '''
        Look up related artist_credits, given a list of artists_credits
    '''

    SERVER_URL = DEVELOPMENT_SERVER_URL + "/artist-credit-similarity/json"

    def __init__(self, threshold=0):
        super().__init__()
        self.threshold = threshold

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):

        artists = inputs[0]
        ac_ids = ",".join([ str(a.artist_credit_id) for a in artists ])
        params = {"[artist_credit_id]": ac_ids,
                  "threshold": self.threshold}

        while True:
            r = requests.get(self.SERVER_URL, params=params)
            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code != 200:
                raise PipelineError("Cannot fetch related artist credits from ListenBrainz: HTTP code %d" % r.status_code)

            break


        try:
            relations = r.text
        except ValueError as err:
            raise PipelineError("Cannot fetch related artist credits from ListenBrainz: Invalid JSON returned: " + str(err))

        index = defaultdict(list)
        for row in relations:
            index[row['artist_credit_id']].append(row)

        entities = []
        for artist in artists:
            a = copy.deepcopy(artist)
            a.mb['related_artist_credit_ids'] = index[artist.artist_credit_id]
            entities.append(a)

        return entities



import unittest
from unittest.mock import patch

from troi import Recording
import troi.listenbrainz.recs

recording_ret = {
    "payload": {
        "count": 3,
        "entity": "recording",
        "last_updated": 1603372967,
        "mbids": [
            {
                "recording_mbid": "5b44c236-cbb0-4e54-a8b1-c344b76e1b8f",
                "score": 1.0
            },
            {
                "recording_mbid": "b6293447-05bf-4123-bfe5-099754348bbf",
                "score": .5
            },
            {
                "recording_mbid": "048e135c-204d-4d77-8e6a-203920768e4f",
                "score": .1
            }
        ],
        "offset": 0,
        "total_mbid_count": 1000,
        "type": "top",
        "user_name": "rob"
    }
}

class TestRecs(unittest.TestCase):

    @patch('liblistenbrainz.ListenBrainz.get_user_recommendation_recordings')
    def test_get_user_recommendation_recordings(self, recs_mock):

        recs_mock.return_value = recording_ret
        u = troi.listenbrainz.recs.UserRecordingRecommendationsElement("rob", artist_type="top", count=3, offset=0)
        entities = u.read()
        recs_mock.assert_called_with("rob", "top", count=3, offset=0)
        assert len(entities) == 3
        assert isinstance(entities[0], Recording)
        assert entities[0].mbid == '5b44c236-cbb0-4e54-a8b1-c344b76e1b8f'
        assert entities[0].ranking == 1.0
        assert entities[1].mbid == 'b6293447-05bf-4123-bfe5-099754348bbf'
        assert entities[1].ranking == .5
        assert entities[2].mbid == '048e135c-204d-4d77-8e6a-203920768e4f'
        assert entities[2].ranking == .1



from time import sleep

import troi
from troi import Recording, Artist
from troi.plist import plist
from troi import TARGET_NUMBER_OF_RECORDINGS
from troi.utils import interleave
from troi.recording_search_service import RecordingSearchByArtistService


class LBRadioArtistRecordingElement(troi.Element):
    """
        Given an artist, find its similar artists and their popular tracks and return one 
        stream of recodings from it.
    """

    MAX_TOP_RECORDINGS_PER_ARTIST = 35  # should lower this when other sources of data get added
    MAX_NUM_SIMILAR_ARTISTS = 8
    MAX_NUM_RECORDINGS = 1000

    def __init__(self, artist_mbid, artist_name, mode="easy", include_similar_artists=True):
        troi.Element.__init__(self)
        self.artist_mbid = str(artist_mbid)
        self.artist_name = artist_name
        self.mode = mode
        self.include_similar_artists = include_similar_artists
        if include_similar_artists:
            self.max_top_recordings_per_artist = self.MAX_TOP_RECORDINGS_PER_ARTIST * 2
        else:
            self.max_top_recordings_per_artist = self.MAX_TOP_RECORDINGS_PER_ARTIST * 3

    def inputs(self):
        return []

    def outputs(self):
        return [Recording]

    def read(self, entities):

        self.data_cache = self.local_storage["data_cache"]

        # Fetch our mode ranges
        start, stop = self.local_storage["modes"][self.mode]
        self.recording_search_by_artist = self.patch.get_service("recording-search-by-artist")
        if self.include_similar_artists:
            similar_artist_count = self.MAX_NUM_SIMILAR_ARTISTS
        else:
            similar_artist_count = 0

        (artist_recordings, msgs) = self.recording_search_by_artist.search(self.mode, self.artist_mbid, start, stop,
                                                                           self.max_top_recordings_per_artist,
                                                                           similar_artist_count)
        # Collect the names of the similar artists
        similar_artist_names = []
        for mbid in artist_recordings:
            if mbid == self.artist_mbid:
                continue

            try:
                similar_artist_names.append(artist_recordings[mbid][0].artist_credit.name)
                # The item may not exist, or the artist credit may be None
            except (AttributeError, IndexError):
                pass

        # craft user feedback messages
        if not artist_recordings:
            msgs.append(f"The seed artist %s has no similar artists, nor top recordings. Too niche?" % self.artist_name)
        else:
            msg = "Using seed artist %s" % self.artist_name
            if self.include_similar_artists:
                if similar_artist_names:
                    msg += " and similar artists: " + ", ".join(similar_artist_names)
                else:
                    msg += " only, since this artist has no similar artists (yet)."
            else:
                msg += " only"

            msgs.insert(0, msg)

        for msg in msgs:
            self.local_storage["user_feedback"].append(msg)
        self.data_cache["element-descriptions"].append("artist %s" % self.artist_name)

        return interleave([artist_recordings[mbid] for mbid in artist_recordings])[:self.MAX_NUM_RECORDINGS]



class Service:
    """
        Services allow troi users to provide local versions of key search functions
        so that both global (e.g. MusicBrainz) and local search contexts can be used.
        
        Downstream users will be able to provide localized version of search functions
        to provide lists of Recordings that meet certain requirements for playlist inclusion.

        Only one service can be registered for any given service slug at a time. The most
        recently registered service will be use for the next playlist generation.
    """

    def __init__(self, slug):
        self._slug = slug

    @property
    def slug(self):
        return self._slug



import mutagen.mp3

from troi.content_resolver.formats.tag_utils import get_tag_value, extract_track_number


EXTENSIONS = {'.mp3', '.mp2', '.m2a'}
READER = mutagen.mp3.MP3


def get_metadata(tags):

    mdata = {}
    if "TPE1" in tags:
        mdata["artist_name"] = str(tags["TPE1"])
    else:
        mdata["artist_name"] = None

    if "TSOP" in tags:
        mdata["sortname"] = str(tags["TSOP"])
    else:
        if "XSOP" in tags:
            mdata["artist_sortname"] = str(tags["XSOP"])
        else:
            mdata["artist_sortname"] = ""

    if "TALB" in tags:
        mdata["release_name"] = str(tags["TALB"])
    else:
        mdata["release_name"] = None

    if "TIT2" in tags:
        mdata["recording_name"] = str(tags["TIT2"])
    else:
        mdata["recording_name"] = None

    if "TRCK" in tags:
        mdata["track_num"] = extract_track_number(str(tags["TRCK"]))
    else:
        mdata["track_num"] = 0

    if "TPOS" in tags:
        mdata["disc_num"] = extract_track_number(str(tags["TPOS"]))
    else:
        mdata["disc_num"] = 1

    if "TXXX:MusicBrainz Artist Id" in tags:
        id = str(tags["TXXX:MusicBrainz Artist Id"])
        # sometimes artist id fields contain two ids. For now, pick the first one and go
        ids = id.split("/")
        mdata["artist_mbid"] = ids[0]
    else:
        mdata["artist_mbid"] = None

    if "UFID:http://musicbrainz.org" in tags:
        mdata["recording_mbid"] = tags["UFID:http://musicbrainz.org"].data.decode("utf-8")
    else:
        mdata["recording_mbid"] = None

    if "TXXX:MusicBrainz Album Id" in tags:
        mdata["release_mbid"] = str(tags["TXXX:MusicBrainz Album Id"])
    else:
        mdata["release_mbid"] = None

    if "TXXX:MusicBrainz Album Artist Id" in tags:
        mdata["release_artist_mbid"] = str(tags["TXXX:MusicBrainz Album Artist Id"])
    else:
        mdata["release_artist_mbid"] = None
    mdata["duration"] = int(tags.info.length * 1000)

    return mdata



import logging
import random
from abc import ABC, abstractmethod
from typing import Dict

from troi.utils import recursively_update_dict

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
_handler = logging.StreamHandler()
_handler.setFormatter(logging.Formatter("%(message)s"))
logger.addHandler(_handler)

DEVELOPMENT_SERVER_URL = "https://datasets.listenbrainz.org"

# Number of recordings that each patch should aim to generate
TARGET_NUMBER_OF_RECORDINGS = 50


class Element(ABC):
    """
        Base class for elements
    """

    def __init__(self, patch=None):
        self.sources = []
        self.patch = patch

    def set_patch_object(self, patch):
        """
            Set patch object -- this is so that each element has a reference to the 
            parent Patch. To keep pipeline construction easy, this is done automatically
            once the pipeline starts executing.
        """
        self.patch = patch

    @property
    def local_storage(self):
        if self.patch is None:
            raise RuntimeError("This Element does not have a patch set. Pass in the patch to the constructor " +
                               "if you need the patch (for local storage) to be available before the pipeline " +
                               "starts executing.")

        return self.patch.local_storage

    def set_sources(self, sources):
        """
           Set the source elements for this element.

           :param sources: A list of objects derived from the Element class, containing at least 1 class. When the pipeline is executed the output of the provided class(es) read function will become the input to this class.
        """

        if not isinstance(sources, list):
            sources = [sources]

        self.sources = sources

        # type check the source
        if len(self.inputs()) > 0:
            for source in sources:
                matched = False if len(source.outputs()) > 0 else True
                for output_type in source.outputs() or []:
                    if output_type in self.inputs():
                        matched = True
                        break

                if not matched:
                    raise RuntimeError("Element %s cannot accept any of %s as input." % (type(self).__name__, self.inputs()))

    def check(self):
        """
            Check to see if the necessary connections are in place
            and that types match correctly.
        """

        if not self.sources:
            raise PipelineError("element %s has no sources defined." % str(type(self)))

        for source in self.sources:
            source.check()

    def generate(self, quiet):
        """
            Generate output from the pipeline. This should be called on
            the last element in the pipeline and no where else. At the root
            node generate returns the results, but on interior nodes it
            returns data to be used as input in the next level.
        """

        source_lists = []
        if self.sources:
            for source in self.sources:
                result = source.generate(quiet)
                if result is None:
                    return None

                if len(self.inputs()) > 0 and \
                    len(result) > 0 and type(result[0]) not in self.inputs() and \
                    len(source.outputs()) > 0:
                    raise RuntimeError("Element %s was expected to output %s, but actually output %s" %
                                       (type(source).__name__, source.outputs()[0], type(result[0])))

                source_lists.append(result)

        items = self.read(source_lists)
        if items is None:
            return None

        if not quiet:
            if len(items) > 0 and type(items[0]) == Playlist:
                logger.info("  %-50s %d items" % (type(self).__name__[:49], len(items[0].recordings or [])))
            else:
                logger.info("  %-50s %d items" % (type(self).__name__[:49], len(items or [])))

        return items

    def run(self):
        """
            This function should be called on the very last element of the
            pipeline to generate output from the timeline. Run will santiy
            check the pipeline first, and then generate output if it can.
        """
        self.check()
        return self.generate()

    @staticmethod
    def inputs():
        """
            Return a list of Artist, ArtistCredit, Release or Recording classes
            that define the type and number of input lists to this element.
            e.g. [ Artist, ArtistCredit, Recording ] means that this element expects
            a list of artist credits and a list of recordings for inputs.
        """
        return None

    @staticmethod
    def outputs():
        """
            Return a list of Artist, ArtistCreditm Release or Recording classes
            that define the type and number of output lists returned by this element.
            e.g. [ Artist, ArtistCredit, Recording ] means that this element returns
            a list of artist credits and a list of recordings.
        """
        return None

    @abstractmethod
    def read(self, source_data_list):
        '''
            This method is where the action happens -- when the consumer wants to
            read data from the pipeline, it calls read() on the last element in
            the pipeline and this casues the while pipeline to generate result.
            If the initializers of other objects in the pipeline are updated,
            calling read() again will generate the set new.

            Note: This function should not be called directly by the user.
        '''

        pass


class Entity(ABC):
    """
        This is the base class for entity objects in troi. Each object will have
        a name, mbid attributes which are all optional. There will also be
        three dicts named musicbrainz, listenbrainz and acousticbrainz that will
        contain collected metadata respective of each project.

        For instance, the musicbrainz dict might have keys that shows the type
        of an artist or the listenbrainz dict might contain the BPM for a track.
        How exactly these dicts will be organized is TDB.
    """

    def __init__(self, mbid=None, ranking=None, musicbrainz=None, listenbrainz=None, acousticbrainz=None):
        self.name = None
        self.mbid = mbid
        self.musicbrainz = musicbrainz or {}
        self.listenbrainz = listenbrainz or {}
        self.acousticbrainz = acousticbrainz or {}
        self.notes = []
        self.ranking = ranking

    @property
    def mb(self):
        return self.musicbrainz

    @property
    def lb(self):
        return self.listenbrainz

    @property
    def ab(self):
        return self.acousticbrainz

    def add_note(self, note):
        '''
            This allows various parts of the pipeline to leave notes/comments in the data/
        '''
        self.notes.append(note)

    def __str__(self):
        return "<Entity()>"


class Area(Entity):
    """
        The class that represents an area.
    """

    def __init__(self, id=id, name=None):
        Entity.__init__(self)
        self.name = name
        self.id = id

    def __str__(self):
        return "<Area('%s', %d)>" % (self.name, self.id)


class Artist(Entity):
    """
        The class that represents an artist.
    """

    def __init__(self,
                 name=None,
                 mbid=None,
                 artist_id=None,
                 join_phrase=None,
                 ranking=None,
                 musicbrainz=None,
                 listenbrainz=None,
                 acousticbrainz=None):
        Entity.__init__(self, mbid=mbid, ranking=ranking, musicbrainz=musicbrainz,
                        listenbrainz=listenbrainz, acousticbrainz=acousticbrainz)
        self.name = name
        self.artist_id = artist_id
        self.join_phrase = join_phrase

    def __str__(self):
        return "<Artist('%s', [%s], %s)>" % (self.name, self.mbid, self.artist_id)


class ArtistCredit(Entity):
    """
        The class that represents an artist credit.
    """

    def __init__(self,
                 name=None,
                 artists=None,
                 artist_credit_id=None,
                 ranking=None,
                 musicbrainz=None,
                 listenbrainz=None,
                 acousticbrainz=None):
        Entity.__init__(self, ranking=ranking, musicbrainz=musicbrainz,
                        listenbrainz=listenbrainz, acousticbrainz=acousticbrainz)
        self.name = name
        self.artists = []
        self.artist_credit_id = artist_credit_id
        if artists:
            if not isinstance(artists, list) and not isinstance(artists, tuple):
                raise TypeError("Artists must be a list.")
            self.artists = artists

    def __str__(self):
        return "<ArtistCredit('%s', [%s], %s)>" % (self.name, ",".join([ a.mbid for a in self.artists ]), self.artist_credit_id)


class Release(Entity):
    """
        The class that represents a release.
    """

    def __init__(self,
                 name=None,
                 mbid=None,
                 artist_credit=None,
                 ranking=None,
                 caa_id=None,
                 caa_release_mbid=None,
                 musicbrainz=None,
                 listenbrainz=None,
                 acousticbrainz=None):
        Entity.__init__(self,
                        ranking=ranking,
                        musicbrainz=musicbrainz,
                        listenbrainz=listenbrainz,
                        acousticbrainz=acousticbrainz)
        self.artist_credit = artist_credit
        self.name = name
        self.mbid = mbid
        self.caa_id = caa_id
        self.caa_release_mbid = caa_release_mbid

    def __str__(self):
        return "<Release('%s', %s)>" % (self.name, self.mbid)


class Recording(Entity):
    """
        The class that represents a recording.
    """

    def __init__(self,
                 name=None,
                 mbid=None,
                 msid=None,
                 duration=None,
                 artist_credit=None,
                 release=None,
                 ranking=None,
                 year=None,
                 spotify_id=None,
                 soundcloud_id=None,
                 apple_music_id=None,
                 musicbrainz=None,
                 listenbrainz=None,
                 acousticbrainz=None):
        Entity.__init__(self, ranking=ranking, musicbrainz=musicbrainz, listenbrainz=listenbrainz, acousticbrainz=acousticbrainz)
        self.duration = duration  # track duration in ms
        self.artist_credit = artist_credit
        self.release = release
        self.name = name
        self.mbid = mbid
        self.msid = msid
        self.year = year
        self.spotify_id = spotify_id
        self.apple_music_id = apple_music_id
        self.soundcloud_id = soundcloud_id

    def __str__(self):
        return "<Recording('%s', %s, %s)>" % (self.name, self.mbid, self.msid)


class Playlist(Entity):
    """
        The class that represents a playlist, which is nothing more than a [Recording] with metadata. This is
        different from the PlaylistElement, which is normally uses to terminate a pipeline for submitting or
        saving result playlists to disk.

        The arguments are the same as the other entities, except that mbid in this case refers to a playlist_mbid
        and that filename is the suggested filename that this playlist should be saved as, if the user asked to 
        do that and didn't provide a different filename.
    """

    def __init__(self,
                 name=None,
                 mbid=None,
                 filename=None,
                 recordings=None,
                 description=None,
                 ranking=None,
                 year=None,
                 musicbrainz=None,
                 listenbrainz=None,
                 acousticbrainz=None,
                 patch_slug=None,
                 user_name=None,
                 additional_metadata=None):
        Entity.__init__(self, ranking=ranking, musicbrainz=musicbrainz, listenbrainz=listenbrainz, acousticbrainz=acousticbrainz)
        self.name = name
        self.filename = filename
        self.mbid = mbid
        self.recordings = recordings
        self.description = description
        self.patch_slug = patch_slug
        self.user_name = user_name
        self.additional_metadata = additional_metadata

    def __str__(self):
        return "<Playlist('%s', %s, %s)>" % (self.name, self.description, self.mbid)

    def add_metadata(self, metadata: Dict):
        """ Update the playlist's additional metadata recursively """
        if self.additional_metadata is None:
            self.additional_metadata = {}

        recursively_update_dict(self.additional_metadata, metadata)

    def shuffle(self, index=None):
        """ Shuffle the playlists randomly, making no effort to make it "nice" for humans. Screw humans
            and their pesky concept of 'random'."""

        random.shuffle(self.recordings)


class User(Entity):
    """
        The class that represents a ListenBrainz user.
    """

    def __init__(self, user_name=None, user_id=None):
        Entity.__init__(self)
        self.user_name = user_name
        self.user_id = user_id

    def __str__(self):
        return "<User('%s', %d)>" % (self.user_name, self.user_id or -1)


class PipelineError(RuntimeError):
    """
        An exception to be thrown when the pipeline encounters an erorr that is a runtime error and not a
        programming error. The main loop will catch this exception and print an error, but not a stacktrace.
    """
    pass



import json

from troi import Element, Artist, Recording, PipelineError

try:
    from openai import OpenAI
except ImportError:
    OpenAI = None

PLAYLIST_PROMPT_PREFIX = "Create a playlist of 50 songs that are suitable for a playlist with the given description:"
PLAYLIST_PROMPT_SUFFIX = "The output should strictly adhere to the following JSON format: the top level JSON object should have three keys, playlist_name to denote the name of the playlist, playlist_description to denote the description of the playlist, and recordings a JSON array of objects where each element JSON object has the recording_name and artist_name keys."


class GPTRecordingElement(Element):
    """ Ask GPT to generate a list of recordings given the playlist description. """

    def __init__(self, api_key, prompt):
        super().__init__()
        if OpenAI is not None:
            self.client = OpenAI(api_key=api_key)
        else:
            raise PipelineError("OpenAI module needs to be installed to use this patch.")
        self.prompt = prompt

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):
        full_prompt = PLAYLIST_PROMPT_PREFIX + " " + self.prompt + " " + PLAYLIST_PROMPT_SUFFIX
        response = self.client.chat.completions.create(
            model="gpt-3.5-turbo-0125",
            response_format={"type": "json_object"},
            messages=[
                {"role": "system", "content": "You are a helpful assistant designed to output JSON."},
                {"role": "user", "content": full_prompt}
            ]
        )

        try:
            playlist = json.loads(response.choices[0].message.content)
        except json.JSONDecodeError as e:
            raise PipelineError("Cannot parse JSON response from OpenAI: %s" % (response.choices[0].message.content,)) \
                from e

        recordings = []
        for item in playlist["recordings"]:
            artist = Artist(name=item["artist_name"])
            recording = Recording(name=item["recording_name"], artist=artist)
            recordings.append(recording)

        self.local_storage["_playlist_name"] = playlist["playlist_name"]
        self.local_storage["_playlist_desc"] = playlist["playlist_description"]

        return recordings



from collections import defaultdict
from time import sleep

import requests
import ujson

from troi import Element, Artist, ArtistCredit, PipelineError, Recording, Playlist, Release


class RecordingLookupElement(Element):
    '''
        Look up a musicbrainz data for a list of recordings, based on MBID.

        :param skip_not_found: If skip_not_found is set to True (the default) then Recordings that cannot be found in MusicBrainz will not be returned from this Element.
    '''

    SERVER_URL = "https://api.listenbrainz.org/1/metadata/recording"
    MAX_RECORDINGS = 1000

    def __init__(self, skip_not_found=True, lookup_tags=False, tag_threshold=None, auth_token=None):
        Element.__init__(self)
        self.skip_not_found = skip_not_found
        self.lookup_tags = lookup_tags
        self.tag_threshold = tag_threshold
        self.auth_token = auth_token

    @staticmethod
    def inputs():
        return [ Recording, Playlist ]

    @staticmethod
    def outputs():
        return [ Recording, Playlist ]

    def read(self, inputs):

        if isinstance(inputs[0], Playlist):
            recordings = inputs[0].recordings
        else:
            recordings = inputs[0]
        if not recordings:
            return []

        recording_mbids = []
        for r in recordings:
            recording_mbids.append(r.mbid)

        if len(recording_mbids) > self.MAX_RECORDINGS:
            raise PipelineError("Cannot fetch more than %d recordings from ListenBrainz." % self.MAX_RECORDINGS)

        inc = "artist release"
        if self.lookup_tags:
            inc += " tag"

        while True:
            headers = {"Authorization": f"Token {self.auth_token}"} if self.auth_token else {}
            r = requests.post(self.SERVER_URL, json={"recording_mbids": recording_mbids, "inc": inc}, headers=headers)
            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code != 200:
                raise PipelineError("Cannot fetch recordings from ListenBrainz: HTTP code %d (%s)" % (r.status_code, r.text))

            break

        try:
            data = ujson.loads(r.text)
        except ValueError as err:
            raise PipelineError("Cannot parse recordings: " + str(err))

        output = []
        for r in recordings:
            # Check if some tracks didn't resolve
            try:
                metadata_recording = data[r.mbid]
            except KeyError:
                if not self.skip_not_found:
                    output.append(r)
                continue

            if self.lookup_tags:
                # Parcel out the tags for artists
                artist_genres = defaultdict(list)
                artist_tags = defaultdict(list)
                for genre in data[r.mbid]["tag"]["artist"]:
                    if "genre_mbid" in genre:
                        artist_genres[genre["artist_mbid"]].append(genre["tag"])
                    else:
                        artist_tags[genre["artist_mbid"]].append(genre["tag"])

            # Now build the artists
            artists = []
            for artist in metadata_recording["artist"]["artists"]:
                artists.append(Artist(mbid=artist["artist_mbid"], 
                                      name=artist["name"],
                                      join_phrase=artist["join_phrase"]))


            # Now that we have artists, we can build artist credits.
            r.artist_credit = ArtistCredit(name=metadata_recording["artist"]["name"],
                                           artists=artists,
                                           artist_credit_id=metadata_recording["artist"]['artist_credit_id'])

            if self.lookup_tags:
                # Finish this, read from dict above
                r.artist_credit.musicbrainz["genre"] = artist_genres[artist["artist_mbid"]]
                r.artist_credit.musicbrainz["tag"] = artist_tags[artist["artist_mbid"]]

            if metadata_recording["release"]:
                # Now create the release data
                r.release = Release(name=metadata_recording["release"]["name"],
                                    mbid=metadata_recording["release"]["mbid"],
                                    caa_id=metadata_recording["release"].get("caa_id", None),
                                    caa_release_mbid=metadata_recording["release"].get("caa_release_mbid", None),
                                    musicbrainz={"release_group_mbid":metadata_recording["release"]["release_group_mbid"]})
            else:
                r.release = None

            if self.lookup_tags:
                # Process the release tags
                genres = []
                tags = []
                for genre in data[r.mbid]["tag"]["release_group"]:
                    if "genre_mbid" in genre:
                        genres.append(genre["tag"])
                    else:
                        tags.append(genre["tag"])

                r.release.musicbrainz = { "genre": genres, "tag": tags }

            # Finally copy data for the recording itself
            r.name = metadata_recording["recording"]['name']
            try:
                r.duration = metadata_recording["recording"]['length']
            except KeyError:
                pass
            try:
                r.year = metadata_recording["release"]["year"]
            except KeyError:
                pass

            if self.lookup_tags:
                # Process the recording tags
                genres = []
                tags = []
                for genre in data[r.mbid]["tag"]["recording"]:
                    if "genre_mbid" in genre:
                        genres.append(genre["tag"])
                    else:
                        tags.append(genre["tag"])

                r.musicbrainz = { "genre": genres, "tag": tags }

            output.append(r)

        if isinstance(inputs[0], Playlist):
            inputs[0].recordings = output
            output = inputs[0]

        return output



import troi
from random import randint, shuffle
from uuid import UUID

import troi
import liblistenbrainz
import liblistenbrainz.errors
from troi import Artist, ArtistCredit, Recording
from troi import TARGET_NUMBER_OF_RECORDINGS
from troi.parse_prompt import TIME_RANGES


class LBRadioStatsRecordingElement(troi.Element):
    """
        Given a LB user, fetch their recording stats and then include recordings from it.
    """

    NUM_RECORDINGS_TO_COLLECT = TARGET_NUMBER_OF_RECORDINGS * 2

    def __init__(self, user_name, time_range, mode="easy"):
        troi.Element.__init__(self)
        self.user_name = user_name
        self.time_range = time_range
        self.mode = mode
        self.client = liblistenbrainz.ListenBrainz()

        if time_range not in TIME_RANGES:
            raise RuntimeError("entity stats must specify one of the following time range options: " + ", ".join(TIME_RANGES))

    def inputs(self):
        return []

    def outputs(self):
        return [Recording]

    def read(self, entities):

        if self.mode == "easy":
            offset = 0
        elif self.mode == "medium":
            offset = 100
        else:
            offset = 200

        # Fetch the user stats
        try:
            result = self.client.get_user_recordings(self.user_name, 100, offset, self.time_range)
        except liblistenbrainz.errors.ListenBrainzAPIException as err:
            raise RuntimeError("Cannot fetch recording stats for user %s" % self.user_name)

        if result is None or "recordings" not in result["payload"]:
            raise RuntimeError("There are no stats available for user '%s' for the %s time_range." %
                               (self.user_name, self.time_range))

        # Give feedback on what we collected
        self.local_storage["data_cache"]["element-descriptions"].append(f"{self.user_name}'s stats for {self.time_range}")

        # Turn them into recordings
        recordings = []
        for r in result['payload']['recordings']:
            if r['recording_mbid'] is not None:
                artists = [Artist(mbid=mbid) for mbid in r["artist_mbids"]]
                artist_credit = ArtistCredit(artists=artists, musicbrainz={"artist_mbids": r["artist_mbids"]})
                recordings.append(Recording(mbid=r['recording_mbid'], artist_credit=artist_credit))

        # Shuffle the recordings
        shuffle(recordings)

        # Check to make sure we're not going to have tracks by the same artist sequentially
        for i in range(len(recordings), 0, -1):
            try:
                if recordings[i].artist_credit.musicbrainz["artist_mbids"] == recordings[
                        i + 1].artist_credit.musicbrainz["artist_mbids"]:
                    recordings.pop(i)
            except IndexError:
                pass

        return recordings



import requests
import ujson
from time import sleep

from troi import Element, Artist, ArtistCredit, Recording, Release, PipelineError


class MBIDMappingLookupElement(Element):
    '''
       Look up MBIDs for the given recordings, if possible. If recordings
       are not found, their data remains unchanged.

       :param remove_unmatched: If True (default False) do not return any Recordings that were not found using the mapping.
    '''

    SERVER_URL = "https://labs.api.listenbrainz.org/mbid-mapping/json"

    def __init__(self, remove_unmatched=False):
        super().__init__()
        self.remove_unmatched = remove_unmatched

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):

        params = []
        for r in inputs[0]:
            if r.artist_credit is not None and r.name is not None:
                params.append({"artist_credit_name": r.artist_credit.name, "recording_name": r.name})

        if not params:
            return []

        while True:
            r = requests.post(self.SERVER_URL, json=params)
            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code != 200:
                raise PipelineError("Cannot fetch MBID mapping rows from ListenBrainz: HTTP code %d (%s)" % (r.status_code, r.text))

            break


        entities = []
        for row in r.json():
            r = inputs[0][int(row['index'])]

            if not row['artist_credit_name']:
                if not self.remove_unmatched:
                    entities.append(r)
                continue

            if r.mbid:
                r.add_note("recording mbid %s overwritten by mbid_lookup" % (r.mbid,))
            r.mbid = row['recording_mbid']
            r.name = row['recording_name']

            r.artist_credit = ArtistCredit(
                artist_credit_id=row['artist_credit_id'],
                name=row['artist_credit_name'],
                artists=[ Artist(mbid=mbid) for mbid in row['artist_mbids'] ] 
            )

            r.release = Release(row['release_name'], mbid=row['release_mbid'])

            entities.append(r)

        return entities



import logging
import os
import json
import hashlib
import mutagen

from troi.content_resolver.model.database import db

logger = logging.getLogger(__name__)


class FindDuplicates:
    '''
       Class to fetch recordings that are duplicate in the database.
    '''

    def __init__(self, db):
        self.db = db

    def get_duplicate_recordings(self, include_different_releases):
        """
           Return a list of (recording_name
        """

        if include_different_releases:
            query = """SELECT recording_name
                            , release_name
                            , artist_name
                            , recording_mbid
                            , json_group_array(file_id) AS file_id
                            , COUNT(*) AS cnt
                         FROM recording
                     GROUP BY recording_mbid
                            , release_mbid
                       HAVING cnt > 1
                     ORDER BY cnt DESC, artist_name, recording_name"""
        else:
            query = """SELECT recording_name
                            , release_name
                            , artist_name
                            , recording_mbid
                            , json_group_array(file_id) AS file_id
                            , COUNT(*) AS cnt
                         FROM recording
                     GROUP BY recording_mbid
                       HAVING cnt > 1
                     ORDER BY cnt DESC, artist_name, recording_name"""

        for r in db.execute_sql(query).fetchall():
            yield (r[0], r[1], r[2], r[3], json.loads(r[4]), r[5])

    @staticmethod
    def sha1sum(filename):
        h = hashlib.sha1()
        mv = memoryview(bytearray(128*1024))
        with open(filename, 'rb', buffering=0) as f:
            while n := f.readinto(mv):
                h.update(mv[:n])
        return h.hexdigest()

    def print_duplicate_recordings(self, include_different_releases=True, verbose=False):

        total = 0
        recordings_count = 0

        def indent(n, s=''):
            return ' ' * (4 * n) + str(s)

        def print_error(e):
            logger.info(indent(2, "error: %s" % e))

        def print_info(title, content):
            logger.info(indent(2, "%s: %s" % (title, content)))

        for dup in self.get_duplicate_recordings(include_different_releases):
            recordings_count += 1
            logger.info("%d duplicates of '%s' by '%s'" % (dup[5], dup[0], dup[2]))
            for file_id in dup[4]:
                logger.info(indent(1, file_id))
                if verbose:
                    error = False
                    try:
                        file_stats = os.stat(file_id)
                        print_info("size", "%d bytes" % file_stats.st_size)
                    except Exception as e:
                        print_error(e)
                        error = True

                    if not error:
                        try:
                            print_info("sha1", self.sha1sum(file_id))
                        except Exception as e:
                            print_error(e)
                            error = True

                    if not error:
                        try:
                            mf = mutagen.File(file_id)
                            print_info("format", mf.info.pprint())
                        except mutagen.MutagenError as e:
                            print_error(e)

                total += 1
            logger.info("")

        logger.info("")
        logger.info("%d recordings had a total of %d duplicates." % (recordings_count, total))
